/* Distributed under the OSI-approved BSD 3-Clause License.  See accompanying
   file Copyright.txt or https://cmake.org/licensing for details.  */
#include "cmVersionComparison.h"

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

namespace cmVersionComparison {
namespace {

int cm_vercmp_dot_decimal(const char* lhss, const char* rhss)
{
  const char* endl = lhss;
  const char* endr = rhss;
  unsigned long lhs, rhs;

  while (((*endl >= '0') && (*endl <= '9')) ||
         ((*endr >= '0') && (*endr <= '9'))) {
    // Do component-wise comparison.
    lhs = strtoul(endl, const_cast<char**>(&endl), 10);
    rhs = strtoul(endr, const_cast<char**>(&endr), 10);

    if (lhs < rhs) {
      return -1;
    }
    if (lhs > rhs) {
      return 1;
    }

    if (*endr == '.') {
      endr++;
    }

    if (*endl == '.') {
      endl++;
    }
  }
  return 0;
}

int cm_vercmp_deb(const char* str1, const char* str2)
{
  int r, c1, c2;

  for (;;) {
    c1 = *str1++;
    c2 = *str2++;
    if (isdigit(c1) && isdigit(c2)) {
      while (c1 == '0') {
        c1 = *str1++;
      }
      while (c2 == '0') {
        c2 = *str2++;
      }
      r = 0;
      while (isdigit(c1) && isdigit(c2)) {
        if (!r) {
          r = c1 - c2;
        }
        c1 = *str1++;
        c2 = *str2++;
      }
      if (isdigit(c1)) {
        return 1;
      }
      if (isdigit(c2)) {
        return -1;
      }
      if (r) {
        return r < 0 ? -1 : 1;
      }
    }
    c1 = c1 == '~' ? -1 : !c1 || isalnum(c1) ? c1 : c1 + 256;
    c2 = c2 == '~' ? -1 : !c2 || isalnum(c2) ? c2 : c2 + 256;
    r = c1 - c2;
    if (r) {
      return r < 0 ? -1 : 1;
    }
    if (!c1) {
      return 0;
    }
  }
}

int cm_vercmp_rpm(const char* str1, const char* str2)
{
  int r = 0;
  const char *end1, *end2;

  for (;;) {
    while (*str1 && !isalnum(*str1) && *str1 != '~') {
      str1++;
    }
    while (*str2 && !isalnum(*str2) && *str2 != '~') {
      str2++;
    }
    if (*str1 && *str1 == '~') {
      if (*str2 && *str2 == '~') {
        str1++;
        str2++;
        continue;
      }
      return -1;
    }
    if (*str2 && *str2 == '~') {
      return 1;
    }
    if (!*str1 || !*str2) {
      break;
    }
    if (isdigit(*str1) || isdigit(*str1)) {
      while (*str1 == '0' && isdigit(str1[1])) {
        str1++;
      }
      while (*str2 == '0' && isdigit(str2[1])) {
        str2++;
      }
      for (end1 = str1; isdigit(*end1);) {
        end1++;
      }
      for (end2 = str2; isdigit(*end2);) {
        end2++;
      }
      r = (end1 - str1) - (end2 - str2);
      if (!r) {
        r = strncmp(str1, str2, end1 - str1);
      }
      if (r) {
        return r > 0 ? 1 : -1;
      }
    } else {
      for (end1 = str1; isalpha(*end1);) {
        end1++;
      }
      for (end2 = str2; isalpha(*end2);) {
        end2++;
      }
      r = (end1 - str1) - (end2 - str2);
      if (r > 0) {
        r = strncmp(str1, str2, end2 - str2);
        return r >= 0 ? 1 : -1;
      }
      if (r < 0) {
        r = strncmp(str1, str2, end1 - str1);
        return r <= 0 ? -1 : 1;
      }
      r = strncmp(str1, str2, end1 - str1);
      if (r) {
        return r > 0 ? 1 : -1;
      }
    }
    str1 = end1;
    str2 = end2;
  }
  return *str1 ? 1 : *str2 ? -1 : 0;
}

typedef int (*cm_vercmp_fun)(const char*, const char*);
bool cm_match(cm_vercmp_fun vercmp_fun, const char* lhss, Operator op,
              const char* rhss)
{
  if (op == AlwaysMatch) {
    return true;
  }
  int ret = vercmp_fun(lhss, rhss);
  switch (op) {
    case Less:
      return ret < 0;
    case LessEqual:
      return ret <= 0;
    case Greater:
      return ret > 0;
    case GreaterEqual:
      return ret >= 0;
    case Equal:
      return ret == 0;
    case NotEqual:
      return ret != 0;
    default:
      break;
  }
  // not reachable
  return false;
}

size_t cm_strverscmp_find_first_difference_or_end(const char* lhs,
                                                  const char* rhs)
{
  size_t i = 0;
  /* Step forward until we find a difference or both strings end together.
     The difference may lie on the null-terminator of one string.  */
  while (lhs[i] == rhs[i] && lhs[i] != 0) {
    ++i;
  }
  return i;
}

size_t cm_strverscmp_find_digits_begin(const char* s, size_t i)
{
  /* Step back until we are not preceded by a digit.  */
  while (i > 0 && isdigit(s[i - 1])) {
    --i;
  }
  return i;
}

size_t cm_strverscmp_find_digits_end(const char* s, size_t i)
{
  /* Step forward over digits.  */
  while (isdigit(s[i])) {
    ++i;
  }
  return i;
}

size_t cm_strverscmp_count_leading_zeros(const char* s, size_t b)
{
  size_t i = b;
  /* Step forward over zeros that are followed by another digit.  */
  while (s[i] == '0' && isdigit(s[i + 1])) {
    ++i;
  }
  return i - b;
}

int cm_strverscmp(const char* lhs, const char* rhs)
{
  size_t const i = cm_strverscmp_find_first_difference_or_end(lhs, rhs);
  if (lhs[i] != rhs[i]) {
    /* The strings differ starting at 'i'.  Check for a digit sequence.  */
    size_t const b = cm_strverscmp_find_digits_begin(lhs, i);
    if (b != i || (isdigit(lhs[i]) && isdigit(rhs[i]))) {
      /* A digit sequence starts at 'b', preceding or at 'i'.  */

      /* Look for leading zeros, implying a leading decimal point.  */
      size_t const lhs_zeros = cm_strverscmp_count_leading_zeros(lhs, b);
      size_t const rhs_zeros = cm_strverscmp_count_leading_zeros(rhs, b);
      if (lhs_zeros != rhs_zeros) {
        /* The side with more leading zeros orders first.  */
        return rhs_zeros > lhs_zeros ? 1 : -1;
      }
      if (lhs_zeros == 0) {
        /* No leading zeros; compare digit sequence lengths.  */
        size_t const lhs_end = cm_strverscmp_find_digits_end(lhs, i);
        size_t const rhs_end = cm_strverscmp_find_digits_end(rhs, i);
        if (lhs_end != rhs_end) {
          /* The side with fewer digits orders first.  */
          return lhs_end > rhs_end ? 1 : -1;
        }
      }
    }
  }

  /* Ordering was not decided by digit sequence lengths; compare bytes.  */
  return lhs[i] - rhs[i];
}

} // namespace

int Compare(std::string const& lhs, std::string const& rhs)
{
  return cm_vercmp_dot_decimal(lhs.c_str(), rhs.c_str());
}

bool Match(std::string const& lhs, Operator op, std::string const& rhs)
{
  return cm_match(cm_vercmp_dot_decimal, lhs.c_str(), op, rhs.c_str());
}

int CompareDEB(std::string const& lhs, std::string const& rhs)
{
  return cm_vercmp_deb(lhs.c_str(), rhs.c_str());
}

bool MatchDEB(std::string const& lhs, Operator op, std::string const& rhs)
{
  return cm_match(cm_vercmp_deb, lhs.c_str(), op, rhs.c_str());
}

int CompareRPM(std::string const& lhs, std::string const& rhs)
{
  return cm_vercmp_rpm(lhs.c_str(), rhs.c_str());
}

bool MatchRPM(std::string const& lhs, Operator op, std::string const& rhs)
{
  return cm_match(cm_vercmp_rpm, lhs.c_str(), op, rhs.c_str());
}

int strverscmp(std::string const& lhs, std::string const& rhs)
{
  return cm_strverscmp(lhs.c_str(), rhs.c_str());
}

} // namespace cmVersionComparison
